;;;-*- Package: TAG-TELEMAN; Syntax: Common-Lisp; Mode: Lisp; Base: 10 -*-

;;; Copyright (c) 1991, 1992 by Xerox Corporation

(cl:defpackage :tag-teleman
  (:use :common-lisp :cl-extensions :string-resource :ssb :trie :tdb
	:vector-lexicon :tag-basics :fsa-tokenizer :tag-analysis
	:class-guesser)
  (:export teleman-tokenizer evaluate-teleman))

(cl:in-package :tag-teleman)

;;; define a lexicon

(defvar *teleman-open-class* '(:idio))

(defclass teleman-lexicon (vector-lexicon suffix-lexicon)
  ()
  (:default-initargs
   :lexicon-open-class *teleman-open-class*
   :lexicon-pathname (tdb-pathname "data" "teleman" "lexicon.recoded")
   :lexicon-size 16000
   :suffix-pathname (tdb-pathname "data" "teleman" "suffix.recoded")))

(def-tokenizer-fsa *teleman-tokenizer-fsa*
  (let ((non-white
	 `(/ #\! #\" #\# #\$ #\% #\& #\' #\( #\) #\* #\+ #\, #\- #\. #\/ #\0
	     #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9 #\: #\; #\< #\= #\> #\? #\@ 
	     #\A #\B #\C #\D #\E #\F #\G #\H #\I #\J #\K #\L #\M #\N #\O #\P
	     #\Q #\R #\S #\T #\U #\V #\W #\X #\Y #\Z #\[ #\\ #\] #\^ #\_ #\` 
	     #\a #\b #\c #\d #\e #\f #\g #\h #\i #\j #\k #\l #\m #\n #\o #\p
	     #\q #\r #\s #\t #\u #\v #\w #\x #\y #\z #\{ #\| #\} #\~))
	(white `(/ #\space #\tab #\page #\newline)))
    `((:word (+ ,non-white))
      (:sent (! (/ #\. #\! #\?) ,white) 1))))

(defvar *teleman-lexicon* nil)

(defclass teleman-tokenizer (basic-tag-ts) ()
  (:default-initargs
    :lexicon (or *teleman-lexicon*
		 (setq *teleman-lexicon* (make-instance 'teleman-lexicon)))
    :tagger-pathname (tdb-pathname "data" "teleman" "teleman")
    :tokenizer-fsa *teleman-tokenizer-fsa*
    :transition-biases
    '((:invalid :- :opro :idio)
      (:invalid :adj :det :opro :intj :idio :gname)
      (:invalid :adv :intj :rpro :opro)
      (:invalid :aux :intj :rpro :idio)
      (:invalid :conj :intj :opro :sent :-)
      (:valid :det :noun :adj :pro :adv :prep :gnoun :presp)
      (:valid :inf :verb :pro :noun :aux :adv :adj :det :gnoun)
      (:invalid :prep :verb :aux :conj)
      (:invalid :num :presp :ppro :opro :name :intj)
      (:invalid :sent :sent :rpro :presp :idio)
      (:invalid :verb :idio :intj))
      
;   :symbol-biases `((:valid ,*teleman-open-class* :idio))
    ))

(setq *tokenizer-class* (find-class 'teleman-tokenizer))

#|
;;;; to train teleman:
(pdefsys:compile-system :tag-trainer)
(pdefsys:load-system :tag-trainer)
(use-package :tag-trainer)
#+excl(setf (sys:gsgc-parameter :generation-spread) 10)
(train-on-files (list (tdb-pathname "data" "teleman" "training.txt"))
		:quantum 10000 :iterations 5)
|#



;;;; Code for Teleman tagger evaluation

(defclass teleman-truth-ts (char-ts)
    ((end-pos :initarg :end :type fixnum)
     (pos :initarg :start :type fixnum)
     (ssb1 :initform (make-ssb))
     (ssb2 :initform (make-ssb))))

(defmethod next-token ((stream teleman-truth-ts))
  (with-slots (char-stream end-pos pos ssb1 ssb2)
      stream
    (loop (multiple-value-bind (token tag end)
	      (parse-token char-stream pos end-pos ssb1 ssb2)
	    (setf pos end)
	    (unless (eq tag :sent)
	      (return (values token tag)))))))

(defun parse-token (stream pos end ssb1 ssb2)
  (declare (fixnum pos end))
  (setf (ssb-pointer ssb1) 0)
  (setf (ssb-pointer ssb2) 0)
  (let ((collecting-tag-p nil))
    (loop (when (>= pos end) (return))
      (let ((char (fast-read-char stream)))
	(incf pos)
	(cond ((eql char #\newline) (return))
	      ((eql char #\space) (setq collecting-tag-p t))
	      (t (if collecting-tag-p
		     (ssb-push-extend char ssb2)
		   (ssb-push-extend char ssb1))))))
    (if collecting-tag-p
	(values (simple-string-copy (ssb-buffer ssb1) (ssb-pointer ssb1))
		(make-tag ssb2)
		pos)
      (values nil nil pos))))		; fell off the end of the file

(defun make-tag (tag-ssb)
  (let* ((string
	  (nstring-upcase
	   (simple-string-copy (ssb-buffer tag-ssb) (ssb-pointer tag-ssb))))
	 (symbol
	  (or (find-symbol string *the-keyword-package*)
	      (intern (copy-seq string) *the-keyword-package*))))
    (free-string string)
    symbol))

      
;;;; Specialized I/O for diagnostics and verification

(defclass teleman-evaluation-ts (tagging-ts)
  ((truth-ts :accessor teleman-truth-ts :initarg :truth-ts)
   (ts-output-p :initarg :output-p :accessor ts-output-p)
   (ts-total :initform 0 :accessor ts-total :type fixnum)
   (ts-ambiguous :initform 0 :accessor ts-ambiguous :type fixnum)
   (ts-errors :initform 0 :accessor ts-errors :type fixnum)
   (ts-confusion-table :initarg :confusion-table :accessor ts-confusion-table)
   (ts-error-trie :initarg :error-trie :accessor ts-error-trie)))

(defmethod next-token ((ts teleman-evaluation-ts))
  (multiple-value-bind (token tag) (call-next-method)
    (multiple-value-bind (token2 true-tag) (next-token (teleman-truth-ts ts))
      (when (and token token2)
	(assert (string= token token2))
	(free-string token2)
	(let ((arity (length (token-class ts))))
	  (incf (ts-total ts))
	  (unless (eq true-tag tag)
	    (incf (ts-errors ts))
	    (let ((table (ts-confusion-table ts))
		  (trie (ts-error-trie ts)))
	      (when table (incf (getf (gethash true-tag table) tag 0)))
	      (when trie (incf (st-get token trie 0)))))
	  (unless (= arity 1) (incf (ts-ambiguous ts)))
	  (values token
		  (when (ts-output-p ts)
		    (format nil "~A~[~*~:;/~D~]~:[*~A~;~*~]"
			    tag (1- arity) arity
			    (eq tag true-tag) true-tag))))))))

(defmacro do-plist ((key val plist) &body body)
  (let ((tail (gensym)))
    `(do* ((,tail ,plist (cddr ,tail))
	   (,key (car ,tail) (car ,tail))
	   (,val (cadr ,tail) (cadr ,tail)))
	 ((null ,tail))
       ,@body)))

(defun evaluate-teleman (source truth &key (output nil)
					 (threshold 100)
					 limit)
  (with-open-file (source-stream source)
    (with-open-file (truth-stream truth)
      (let ((ts
	     (make-instance 'teleman-evaluation-ts
	       :char-stream source-stream
	       :truth-ts (make-instance 'teleman-truth-ts
			   :char-stream truth-stream)
	       :output-p (and output t)
	       :confusion-table (and threshold (make-hash-table :test 'eq))
	       :error-trie (and threshold (make-string-trie)))))
	(if (ts-output-p ts)
	    (print-token-stream ts output limit)
	  (time (loop (free-string (or (next-token ts) (return)))
		  (when (and limit (> (ts-total ts) limit)) (return))
		  (when (zerop (mod (ts-total ts) 1000)) (format t ".")))))
	(format t "~%    total: ~7:D (~4F% correct)~%"
		(ts-total ts)
		(- 100.0 (/ (* (ts-errors ts) 100.0) (ts-total ts))))
	(format t "ambiguous: ~7:D (~4F% correct)~%"
		(ts-ambiguous ts)
		(- 100.0 (/ (* (ts-errors ts) 100.0) (ts-ambiguous ts))))
	(when (ts-confusion-table ts)
	  (format t "Errors by tag, cut at ~D errors:~%" threshold)
	  (let ((*print-case* :downcase))
	    (format t "~:{~4D ~A~%~:{  ~4D ~A~%~}~}"
		    (sort
		     (with-collection
		       (maphash
			#'(lambda (tag errors)
			    (let* ((sum 0)
				   (errors 
				    (with-collection
				      (do-plist (tag count errors)
					(incf sum count)
					(when (>= count threshold)
					  (collect (list count tag)))))))
			      (when (>= sum threshold)
				(collect 
				 (list sum tag
				       (sort errors #'> :key #'car))))))
			(ts-confusion-table ts)))
		     #'> :key #'car))))
	(when threshold
	  (format t "Terms with more than ~D errors:~%" threshold)
	  (format t "~:{~4D ~A~%~}"
		  (sort (with-collection
			  (map-st
			   #'(lambda (term freq)
			       (when (>= freq threshold)
				 (collect (list freq (copy-seq term)))))
			   (ts-error-trie ts)))
			#'> :key #'car)))))))
  
